// NavigateCalculations.c
//
// support file for Navigate
//
// This code is freely distributable. Please contact the author with
// bug reports or any questions regarding the navigation calculations.
//
// Written by: Rick Chapman
//             JHU/APL
//             rick.chapman@jhuapl.edu
//			   8 November 2002
//
// This program was developed on a Macintosh with Code Warrior for Palm OS v 8.0


#include <math.h>
#include "NavigateCalculations.h"

//#pragma warn_stack_usage 4096

// ********************************************************************
// Entry Points
// ********************************************************************

// ********************************************************************
// Global variables
// ********************************************************************

//double  nan;
double g_ra[] = {6377563.396,6378160,6377397.155,6377483.865,6378206.4,6378249.145,
				 6377298.556,6377276.345,6377301.243,6377309.613,6377304.063,6377295.664,
				 6378137,6378200,6378270,6378160,6378388,
				 6378245,6377340.189,6378155,6378160,6378135,6378137};
double g_invf[] = {299.3249646,298.25,299.1528128,299.1528128,294.9786982,293.465,
				300.8017,300.8017,300.8017,300.8017,300.8017,300.8017,
				298.257222101,298.3,297,298.247,297,
				298.3,299.3249646,298.3,298.25,298.26,298.257223563};


// ********************************************************************
// Navigation Calculation Functions
// ********************************************************************


// find geodesic range and initial bearing from loc1 to loc2
// lat and lng are given in degrees, with negative values being west of the prime meridian and south of the equator
// bearing is in degrees true, range is in meters
// uses datum specified in g_ra[22] and g_prefs.invf
void GeoRangeAndBearing(double lat1, double lng1, double lat2, double lng2, double *range, double *bearing) {
	double f = 1.0/g_invf[22];	// Flattening factor.
	double ra = g_ra[22];			// Semimajor axis (m).
	double eps=5.e-14;
	double radeg = 180.0/pi;		// Must use double for mm accuracy.
	
	double r=1.-f;
	double tu1=r*tan(lat1/radeg);
	double tu2=r*tan(lat2/radeg);
	double cu1=1./sqrt(tu1*tu1+1.);
	double su1=cu1*tu1;
	double cu2=1./sqrt(tu2*tu2+1.);
	double s=cu1*cu2;
	double baz=s*tu2;
	double faz=baz*tu1;
	double x=(lng2-lng1)/radeg;
	
	double d,sx,cx,sy,cy,cz,c2a,c,e,y,sa;
	do{
		sx=sin(x);
		cx=cos(x);
		tu1=cu2*sx;
		tu2=baz-su1*cu2*cx;
		sy=sqrt(tu1*tu1+tu2*tu2);
		cy=s*cx+faz;
		y=atan2(sy,cy);
		sa=s*sx/sy;
		c2a=-sa*sa+1.;
		cz=faz+faz;
		if(c2a>0.) cz=-cz/c2a+cy;
		e=cz*cz*2.-1.;
		c=((-3.*c2a+4.)*f+4.)*c2a*f/16.;
		d=x;
		x=((e*cy*c+cz)*sy*c+y)*sa;
		x=(1.-c)*x*f+(lng2-lng1)/radeg;
	}while(fabs(d-x)>eps);
	faz=atan2(tu1,tu2);
	baz=atan2(cu1*sx,baz*cx-su1*cu2)+pi;
	x=sqrt((1/r/r-1.)*c2a+1.)+1.;
	x=(x-2.)/x;
	c=1.-x;
	c=(x*x/4.+1.)/c;
	d=(0.375*x*x-1.)*x;
	x=e*cy;
	s=1.-e*e;
	s=((((sy*sy*4.-3.)*s*cz*d/6.-x)*d/4.+cz)*sy*d+y)*c*ra*r;
	
	*range = s;
	*bearing = faz*radeg;
	if( *bearing > 360.0 ) *bearing -= 360.0;
	if( *bearing < 0.0 ) *bearing += 360.0;
	return;
}

// find end location of geodesic segment
// lat and lng are given in degrees, with negative values being west of the prime meridian and south of the equator
// bearing is in degrees true, range is in meters
// uses datum specified in g_ra[22] and g_invf[22]
void GeoLine(double lat1, double lng1, double bearing, double range, double *lat2, double *lng2) {
	double f = 1.0/g_invf[22];	// Flattening factor.
	double ra = g_ra[22];			// Semimajor axis (m).
	double eps=1.e-13;
	double degrad = pi/180.0;		// Must use double for mm accuracy.
	double r=1.-f;
	double tu=r*tan(lat1*degrad);
	double sf=sin(bearing*degrad);
	double cf=cos(bearing*degrad);
	double baz=atan2(tu,cf)*2.;
	double cu=1./sqrt(tu*tu+1.);
	double su=tu*cu;
	double sa=cu*sf;
	double c2a=1.-sa*sa;
	double x=sqrt((1./r/r-1.)*c2a+1.)+1.;
	double c,d,y,sy,cy,cz,e;
	
	x=(x-2.)/x;
	c=1.-x;
	c=(x*x/4.+1.)/c;
	d=(0.375*x*x-1.)*x;
	tu=range/r/ra/c;
	y=tu;
	do{
		sy=sin(y);
		cy=cos(y);
		cz=cos(baz+y);
		e=cz*cz*2.-1.;
		c=y;
		x=e*cy;
		y=e*e-1.;
		y=(((sy*sy*4.-3.)*y*cz*d/6.+x)*d/4.-cz)*sy*d+tu;
	}while(fabs(y-c)>eps);
	baz=cu*cy*cf-su*sy;
	c=r*sqrt(sa*sa+baz*baz);
	d=su*cy+cu*sy*cf;
	*lat2=atan2(d,c)/degrad;
	c=cu*cy-su*sy*cf;
	x=atan2(sy*sf,c);
	c=((-3.*c2a+4.)*f+4.)*c2a*f/16.;
	d=((e*cy*c+cz)*sy*c+y)*sa;
	*lng2 = (lng1*degrad+x-(1.-c)*d*f)/degrad;
	baz=atan2(sa,baz)+pi;
	
	*lng2 = fmod( *lng2, 360.0 );
	if( *lng2 > 180.0 )
		*lng2 -= 360.0;
	if( *lng2 <= -180.0 )
		*lng2 += 360.0;
	
	return;
}


// find loxodromic range and initial bearing from loc1 to loc2
// lat and lng are given in degrees, with negative values being west of the prime meridian and south of the equator
// bearing is in degrees true, range is in meters
// uses datum specified in g_ra[22] and g_invf[22]
void LoxoRangeAndBearing(double lat1, double lng1, double lat2, double lng2, double *range, double *bearing) {
	double f = 1.0/g_invf[22];	// Flattening factor.
	double a = g_ra[22];			// Semimajor axis (m).
	double b = a*(1-f);				// Semiminor axis (m).
	double e2 = (a*a-b*b)/(a*a);	// Eccentricity^2.
	double e4 = e2*e2;
	double e6 = e2*e4;
	double e = sqrt(e2);			// Eccentricity.
	double radeg = 180.0/pi;		// Must use double for mm accuracy.
//	double pi2 = pi/2.0;

	double lt1,ln1,lt2,ln2,snlat,t1,t2,x1,y1;
	double x2,y2,azir,azi,tt,aa,bb,cc,m1,m2,caz;
	
	if( fabs(lng2-lng1) > 180.0 ) {
		if( lng1 < 0.0 )
			lng1 += 360.0;
		else if( lng2 < 0.0 )
			lng2 += 360.0;
		else if( lng1 > lng2 )
			lng1 -= 360.0;
		else
			lng2 -= 360.0;
	}
	
	lt1 = lat1/radeg;		// Convert long,lat to radians.
	ln1 = lng1/radeg;

	lt2 = lat2/radeg; 		// Convert coordinates to radians.
	ln2 = lng2/radeg;
	
	snlat = sin(lat1/radeg);
	t1 = (1.0+snlat)/(1.0-snlat);
	t2 = (1.0-e*snlat)/(1.0+e*snlat);
	x1 = a*lng1/radeg;
	y1 = (a/2.0)*log(t1*pow(t2,e));

	snlat = sin(lat2/radeg);
	t1 = (1.0+snlat)/(1.0-snlat);
	t2 = (1.0-e*snlat)/(1.0+e*snlat);
	x2 = a*lng2/radeg;
	y2 = (a/2.0)*log(t1*pow(t2,e));

	azir = atan2(x2-x1,y2-y1);			// Azi in radians (7-14).
	azi = azir*radeg;					// Azimuth in degrees.
	
	if ( lat1 == lat2 ) {				// Dist (7-17)
		*range = fabs(a*(ln2-ln1)*cos(lt1)/sqrt(1.0-e2*pow(sin(lt1),2)));
	} else {							//(3-21)
		tt = 1 - e2/4 - 3*e4/64 - 5*e6/256;
		aa = 3*e2/8 + 3*e4/32 + 45*e6/1024;
		bb = 15*e4/256 + 45*e6/1024;
		cc = 35*e6/3072;
		m1 = a*(tt*lt1 - aa*sin(2*lt1) + bb*sin(4*lt1) - cc*sin(6*lt1));
		m2 = a*(tt*lt2 - aa*sin(2*lt2) + bb*sin(4*lt2) - cc*sin(6*lt2));
		caz = cos(azir);
		*range = (m2-m1)/caz;			// (7-16)
	}
	*bearing = azi;
	if( *bearing > 360.0 ) *bearing -= 360.0;
	if( *bearing < 0.0 ) *bearing += 360.0;

	return;
}


//find end location of loxodromic segment
// lat and lng are given in degrees, with negative values being west of the prime meridian and south of the equator
// bearing is in degrees true, range is in meters
// uses datum specified in g_ra[22] and g_invf[22]
void LoxoLine(double lat1, double lng1, double bearing, double range, double *lat2, double *lng2) {
	double f = 1.0/g_invf[22];	// Flattening factor.
	double a = g_ra[22];			// Semimajor axis (m).
	double b = a*(1-f);				// Semiminor axis (m).
	double e2 = (a*a-b*b)/(a*a);	// Eccentricity^2.
	double e4 = e2*e2;
	double e6 = e2*e4;
	double e = sqrt(e2);			// Eccentricity.
	double radeg = 180.0/pi;		// Must use double for mm accuracy.
//	double pi2 = pi/2.0;
	double eps = 0.01;

	double lt1 = lat1/radeg;		// Convert long,lat to radians.
	double ln1 = lng1/radeg;

	double ln2,p,tt,aa,bb,cc,dd,m1,caz,m2,u,tmp,e1,e12,e13,e14,lt2,esin,t1,t2,snlat,y1;
	char specialCase;
	
	bearing = fmod( bearing, 360.0 );
	if( bearing < 0.0 )
		bearing += 360.0;

	specialCase = (fabs(bearing - 90) < eps) || (fabs(bearing - 270) < eps);	// Nearly due east or west
	
	p = lt1;
	tt = 1 - e2/4 - 3*e4/64 - 5*e6/256; 			// (3-21)
	aa = 3*e2/8 + 3*e4/32 + 45*e6/1024;
	bb = 15*e4/256 + 45*e6/1024;
	cc = 35*e6/3072;
	m1 = a*(tt*p - aa*sin(2*p) + bb*sin(4*p) - cc*sin(6*p));
	caz = cos(bearing/radeg);
	m2 = range * caz + m1;							// (7-18)
	
	u = m2/(a*(1-e2/4-3*e4/64-5*e6/256));			// (7-19)
	tmp = sqrt(1-e2);								// (3-24)
	e1 = (1-tmp)/(1+tmp);
	e12 = e1*e1;									// (3-26)
	e13 = e1*e12;
	e14 = e1*e13;
	aa = 3*e1/2 - 27*e13/32;
	bb = 21*e12/16 - 55*e14/32;
	cc = 151*e13/96;
	dd = 1097*e14/512;
	lt2 = u + aa*sin(2*u) + bb*sin(4*u) + cc*sin(6*u) + dd*sin(8*u);
	*lat2 = lt2*radeg;

	if( specialCase ) {								// Nearly due east or west
		ln2 = ln1 + range * sqrt(1-e2*pow(sin(lt1),2))/(a*cos(lt1)) * sin(bearing/radeg);	// (modified 7-17)
		*lng2 = ln2*radeg;
	} else {										// Any other azi.
		snlat = sin(lat1/radeg);
		t1 = (1.0+snlat)/(1.0-snlat);
		t2 = (1.0-e*snlat)/(1.0+e*snlat);
		y1 = (a/2.0)*log(t1*pow(t2,e));
		
		esin = e*sin(lt2);
		t1 = pow(((1-esin)/(1+esin)),e/2);
		t2 = tan(pi/4 + lt2/2);
		ln2 = ln1 + tan(bearing/radeg)*(log(t2*t1) - y1/a);	// (7-20)
		*lng2 = ln2*radeg;
	}

	*lng2 = fmod( *lng2, 360.0 );
	if( *lng2 > 180.0 )
		*lng2 -= 360.0;
	if( *lng2 <= -180.0 )
		*lng2 += 360.0;

	return;
}

// Compute maximum loxodrome length to get to the pole
// lat and lng are given in degrees, with negative values being west of the prime meridian and south of the equator
// bearing is in degrees true, range is in meters
// uses datum specified in g_ra[22] and g_invf[22]
void MaxLoxoLine(double lat1, double lng1, double bearing, double *range, double *lat2, double *lng2) {
	double f = 1.0/g_invf[22];	// Flattening factor.
	double a = g_ra[22];			// Semimajor axis (m).
	double b = a*(1-f);				// Semiminor axis (m).
	double e2 = (a*a-b*b)/(a*a);	// Eccentricity^2.
	double e4 = e2*e2;
	double e6 = e2*e4;
	double e = sqrt(e2);			// Eccentricity.
	double radeg = 180.0/pi;		// Must use double for mm accuracy.
//	double pi2 = pi/2.0;

	double lt1,ln1,lt2,ln2,snlat,t1,t2,x1,y1;
	double x2,y2,azir,tt,aa,bb,cc,m1,m2,caz;

	*lat2 = 90.0;
	*lng2 = 0.0;

	lt1 = lat1/radeg;		// Convert long,lat to radians.
	ln1 = lng1/radeg;

	lt2 = *lat2/radeg; 		// Convert coordinates to radians.
	ln2 = *lng2/radeg;
	
	snlat = sin(lat1/radeg);
	t1 = (1.0+snlat)/(1.0-snlat);
	t2 = (1.0-e*snlat)/(1.0+e*snlat);
	x1 = a*lng1/radeg;
	y1 = (a/2.0)*log(t1*pow(t2,e));

	snlat = sin(*lat2/radeg);
	t1 = (1.0+snlat)/(1.0-snlat);
	t2 = (1.0-e*snlat)/(1.0+e*snlat);
	x2 = a*(*lng2/radeg);
	y2 = (a/2.0)*log(t1*pow(t2,e));

	azir = bearing/radeg;			// Azi in radians (7-14).
	
	if ( lat1 == *lat2 ) {				// Dist (7-17)
		if( lat1 == 90.0 || lat1 == -90.0 )
			*range = 0.0;
		//else
		//	*range = 1.0 / 0.0;
	} else {							//(3-21)
		tt = 1 - e2/4 - 3*e4/64 - 5*e6/256;
		aa = 3*e2/8 + 3*e4/32 + 45*e6/1024;
		bb = 15*e4/256 + 45*e6/1024;
		cc = 35*e6/3072;
		m1 = a*(tt*lt1 - aa*sin(2*lt1) + bb*sin(4*lt1) - cc*sin(6*lt1));
		m2 = a*(tt*lt2 - aa*sin(2*lt2) + bb*sin(4*lt2) - cc*sin(6*lt2));
		caz = cos(azir);
		*range = (m2-m1)/caz;			// (7-16)
	}
	
	return;
}

